/*
 * Simple FDD bad sector check
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <iocs.h>
#include <string.h>

#define bool unsigned char
#define false                       0
#define true                        1

#define COLOR_BLACK                 4
#define COLOR_BLUE                  5
#define COLOR_YELLOW                6
#define COLOR_WHITE                 7

#define FD2HD0                      0x90
#define FD2HD1                      0x91
#define FD2HD2                      0x92
#define FD2HD3                      0x93
#define MFM_RETRY_SEEK              0x70
#define MFM_NORETRY_SEEK            0x50

#define STATUS_OK                   0x00
#define STATUS_MASK                 0xF8000000
#define STATUS_NO_DISK              0x68
#define STATUS_NO_DRIVE             0x48
#define STATUS_BAD_SECTOR           0x40

#define PROGRESS_WINDOW_X           9
#define PROGRESS_WINDOW_Y           5
#define MESSAGE_WINDOW_X            88
#define MESSAGE_WINDOW_Y            0
#define FDD_WINDOW_X                9
#define FDD_WINDOW_Y                0

#define KEY_SPACE                   32
#define KEY_ENTER                   13
#define KEY_RIGHT                   83
#define KEY_LEFT                    8
#define KEY_R                       114
#define KEY_Q                       113

#define NUMER_OF_SECTORS            1216 // STANDARD X68000 2HD floppy

struct SectorStatus
{
    char status;
    char track;
    char side;
    char sector;
};

struct CheckDiskVariables
{
    bool quit;
    bool stop;
    bool startscan;
    bool newscan;
    bool rescan;
    int fddChanged;
    int fdd;
    int side;
    int track;
    int sector;
    int key;
    struct SectorStatus sectorStatus[NUMER_OF_SECTORS];
};


int Diagnose(unsigned char pda, unsigned char mode, unsigned char track, unsigned char side,  unsigned char sector)
{
    unsigned char data[1024];
    int disk = (pda<<8)|mode;
    int addr = (0x03<<24)|(track<<16)|(side<<8)|sector;
    int status;

    status = _iocs_b_readdi(disk, addr, sizeof(data), data);
    status &= STATUS_MASK;
    status >>= 24;

    return status;
}

void Print(int x, int y, void *text, int color)
{
    _dos_c_locate(x,y);
    _dos_c_color(color);
    _dos_c_print(text);
}

void CLS()
{
    _dos_c_cls_al();
}

void RenderProgressWindow(int x, int y)
{   char row[] = "[*****************************************************************************]";
    int dx, dy, sector, side;

    for (side = 0; side < 2; side++)
    {
        for (sector = 1; sector <= 8; sector++)
        {
            dx = 0;
            dy = (side * 9) + sector - 1;
            Print(x+dx, y+dy, row, COLOR_WHITE);
        }
    }
}

void RenderFDDWindow(int pick)
{
    char fdd[10];
    int i, dx, color;

    for (i = 0; i < 4; i++)
    {
        dx = i*7;
        if (FD2HD0+i == pick)
            color = COLOR_YELLOW;
        else
            color = COLOR_WHITE;

        snprintf(fdd, sizeof(fdd), "[2HD%i]", i);
        Print(FDD_WINDOW_X+dx, FDD_WINDOW_Y, fdd, color);
    }
}

void RenderMessageWindow(char *msg, int color)
{
    char blank[] = "                                            ";

    Print(MESSAGE_WINDOW_X-strlen(blank), MESSAGE_WINDOW_Y, blank, COLOR_BLACK);
    Print(MESSAGE_WINDOW_X-strlen(msg), MESSAGE_WINDOW_Y, msg, color);
}

bool RenderProgressStatus(int track, int side, int sector, int status)
{
    char text[30];

    switch(status)
    {
        case STATUS_OK:
        {
            Print(PROGRESS_WINDOW_X + track + 1, PROGRESS_WINDOW_Y + (side * 9) + sector -1, "O", COLOR_BLUE);
        }
        break;

        case STATUS_BAD_SECTOR:
        {
            Print(PROGRESS_WINDOW_X + track + 1, PROGRESS_WINDOW_Y + (side * 9) + sector -1, "X", COLOR_YELLOW);
        }
        break;

        case STATUS_NO_DISK:
        {
            RenderMessageWindow("[NO DISK!]", COLOR_WHITE);
            return(true);
        }
        break;

        case STATUS_NO_DRIVE:
        {
            RenderMessageWindow("[NO DRIVE!]", COLOR_WHITE);
            return(true);
        }
        break;

        default:
        {
            snprintf(text, sizeof(text), "[UNKNOWN ERROR: %i]", status);
            RenderMessageWindow(text, COLOR_WHITE);
            return(true);
        }
        break;
    }
    return(false);
}

void CheckKey(struct CheckDiskVariables *v)
{
    v->key = _dos_k_keyinp();

    if (!v->key)
        return;

    if (v->key == KEY_LEFT)
    {
        v->fdd--;
        if (v->fdd < FD2HD0) v->fdd = FD2HD0;
        else v->fddChanged = true;

        RenderFDDWindow(v->fdd);
        return;
    }

    if (v->key == KEY_RIGHT)
    {
        v->fdd++;
        if (v->fdd > FD2HD3) v->fdd = FD2HD3;
        else v->fddChanged = true;

        RenderFDDWindow(v->fdd);
        return;
    }

    if (v->key == KEY_ENTER)
    {
        v->stop = true;
        v->startscan = true;
        v->newscan = true;
        v->rescan = false;
        return;
    }

    if (v->key == KEY_Q)
    {
        v->quit = true;
        v->stop = true;
        v->startscan = false;
        v->newscan = false;
        v->rescan = false;
        return;
    }

    if (v->newscan)
        return;

    if (v->key == KEY_SPACE)
    {
        if (v->stop)
        {
            v->startscan = true;
            v->stop = false;
        }
        else
        {
            v->startscan = false;
            v->stop = true;
        }

        v->newscan = false;
        RenderMessageWindow("[PAUSED]", COLOR_WHITE);
        return;
    }

    if (v->key == KEY_R)
    {
        v->stop = true;
        v->startscan = true;
        v->newscan = false;
        v->rescan = true;
        return;
    }
}

void CheckDisk()
{
    char msg[20];
    int i;
    int status;
    struct CheckDiskVariables v;

    v.quit = false;
    v.stop = true;
    v.startscan = false;
    v.newscan = true;
    v.rescan = false;
    v.fdd = FD2HD0;
    v.fddChanged = false;
    v.track = 0;
    v.side = 0;
    v.sector = 1;
    CLS();
    _dos_c_curoff();
    RenderFDDWindow(v.fdd);
    RenderMessageWindow("[PRESS ENTER TO START SCAN]", COLOR_WHITE);
    RenderProgressWindow(PROGRESS_WINDOW_X, PROGRESS_WINDOW_Y);

    while (!v.quit)
    {
        if (v.startscan)
        {
            v.startscan = false;
            v.stop = false;

            if (v.fddChanged)
            {
                v.rescan = false;
                v.newscan = true;
                v.fddChanged = false;
            }

            if (v.rescan)
            {
                RenderMessageWindow("[RESCANNING BAD SECTORS]", COLOR_WHITE);

                for (i = 0; i < NUMER_OF_SECTORS; i++)
                {
                    if (v.sectorStatus[i].status == -1)
                        break;

                    status = Diagnose(v.fdd, MFM_RETRY_SEEK, v.sectorStatus[i].track, v.sectorStatus[i].side, v.sectorStatus[i].sector);
                    v.stop = RenderProgressStatus(v.sectorStatus[i].track, v.sectorStatus[i].side, v.sectorStatus[i].sector, status);
                    if (!v.stop)
                        CheckKey(&v);

                     if (v.stop)
                        break;
                }

                if (!v.stop)
                {
                    if (i == 0)
                        RenderMessageWindow("[NO BAD SECTORS TO SCAN]", COLOR_WHITE);
                    else
                        RenderMessageWindow("[DONE]", COLOR_WHITE);
                    v.rescan = false;
                }
            }
            else
            {
                if (v.newscan)
                {
                    v.newscan = false;
                    CLS();
                    RenderFDDWindow(v.fdd);
                    RenderProgressWindow(PROGRESS_WINDOW_X, PROGRESS_WINDOW_Y);
                    v.track = 0;
                    v.side = 0;
                    v.sector = 1;
                    for (i = 0; i < NUMER_OF_SECTORS; i++)
                        v.sectorStatus[i].status = -1;
                }

                RenderMessageWindow("[SCANNING]", COLOR_WHITE);

                for (; v.track < 77; v.track++)
                {
                    if (v.side > 1)
                        v.side = 0;
                    for (; v.side < 2; v.side++)
                    {
                        if (v.sector > 8)
                            v.sector = 1;
                        for (; v.sector <= 8; v.sector++)
                        {
                            status = Diagnose(v.fdd, MFM_RETRY_SEEK, v.track, v.side, v.sector);
                            v.stop = RenderProgressStatus(v.track, v.side, v.sector, status);

                            if (!v.stop)
                            {
                                if (status == STATUS_BAD_SECTOR)
                                {
                                    for (i = 0; i < NUMER_OF_SECTORS; i++)
                                        if (v.sectorStatus[i].status == -1) break;

                                    v.sectorStatus[i].status = status;
                                    v.sectorStatus[i].track = v.track;
                                    v.sectorStatus[i].side = v.side;
                                    v.sectorStatus[i].sector = v.sector;
                                }
                                CheckKey(&v);
                                if (!v.stop)
                                {
                                    snprintf(msg, sizeof(msg), "T:%i Si:%i Se:%i", v.track,v.side,v.sector);
                                    RenderMessageWindow(msg, COLOR_YELLOW);
                                }
                            }
                            if (v.stop)
                                break;
                        }
                        if (v.stop) break;
                    }
                    if (v.stop) break;
                }

                if (!v.stop)
                {
                    RenderMessageWindow("[DONE]", COLOR_WHITE);
                }
            }
        }
        CheckKey(&v);
    }
}

void Cleanup()
{
    Print(0, 24, "FDR68K v1.0: May the gaming gods be with you.", COLOR_YELLOW);
    _dos_c_color(3);
    _dos_c_locate(0, 26);
    _dos_c_curon();
}
int main (int argc, char **argv)
{
    CheckDisk();
    Cleanup();
    return 0;
}
